// The source
#include <vtkParametricRandomHills.h>
#include <vtkParametricFunctionSource.h>
#include <vtkPolyData.h>
#include <vtkPointData.h>
#include <vtkSphereSource.h>
#include <vtkPlaneSource.h>
#include <vtkElevationFilter.h>
// Curvatures
#include <vtkClipPolyData.h>
#include <vtkCurvatures.h>
#include <vtkPlane.h>
#include <vtkImplicitBoolean.h>
// For annotating
#include <vtkVariantArray.h>
// Lookup table
#include <vtkColorSeries.h>
#include <vtkLookupTable.h>
// For glyphing
#include <vtkReverseSense.h>
#include <vtkMaskPoints.h>
#include <vtkArrowSource.h>
#include <vtkGlyph3D.h>
// For contouring
#include <vtkBandedPolyDataContourFilter.h>
// Mappers, actors, renderers etc.
#include <vtkPolyDataMapper.h>
#include <vtkActor.h>
#include <vtkProperty.h>
#include <vtkScalarBarActor.h>
#include <vtkRenderer.h>
#include <vtkCamera.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>

#include <vtkSmartPointer.h>
#define VTK_SP(type, name)\
  vtkSmartPointer<type> name = vtkSmartPointer<type>::New()

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include <functional>
#include <iterator>
#include <cstdlib>
#include <cctype>
#include <cmath>

namespace {
  //! Some STL Utilities.
  class STLHelpers
  {
  public:
    //---------------------------------------------------------------------------
    STLHelpers(){}

    //---------------------------------------------------------------------------
    virtual ~STLHelpers(){}

    //-----------------------------------------------------------------------------
    // Convert a string to lowercase.
    std::string ToLowercase(const std::string & str)
    {
      std::string s;
      std::transform(str.begin(), str.end(), std::back_inserter(s),
        (int(*)(int))std::tolower);
      return s;
    }

    //-----------------------------------------------------------------------------
    // Replace all occurrences of old_value in a string with new_value.
    std::string ReplaceAll(std::string & str,
      const std::string & old_value, const std::string & new_value)
    {
      size_t start_pos = 0;
      while ((start_pos = str.find(old_value, start_pos)) != std::string::npos)
      {
        str.replace(start_pos, old_value.length(), new_value);
        // It could be that 'new_value' is a substring of 'old_value'.
        start_pos += new_value.length();
      }
      return str;
    }

    //-----------------------------------------------------------------------------
    // An implementation of the C++11 next(iter,n) found in the header <iterator>.
    // ForwardIt must meet the requirements of ForwardIterator.
    // Return the nth successor of iterator it.
    template <typename ForwardIt>
    ForwardIt Next(ForwardIt iter,
      typename std::iterator_traits<ForwardIt>::difference_type n = 1)
    {
      std::advance(iter, n);
      return iter;
    }

    //-----------------------------------------------------------------------------
    // Return true if the iterator points to the last element.
    template <typename Iter, typename Cont>
    bool IsLast(Iter iter, const Cont & cont)
    {
      return (iter != cont.end()) && (Next(iter) == cont.end());
    }

  };
}

//-----------------------------------------------------------------------------
// Function declarations.

//! Divide a range into bands
/*!
@param dR - [min, max] the range that is to be covered by the bands.
@param numberOfBands - the number of bands, a positive integer.
@param nearestInteger - if True then [floor(min), ceil(max)] is used.
@return A List consisting of [min, midpoint, max] for each band.
*/
std::vector<std::vector<double> > MakeBands(double const dR[2],
  int const & numberOfBands, bool const & nearestInteger);

//! Divide a range into integral bands
/*!
Divide a range into bands
@param dR - [min, max] the range that is to be covered by the bands.
@return A List consisting of [min, midpoint, max] for each band.
*/
std::vector<std::vector<double> > MakeIntegralBands(double const dR[2]);

//! Print the bands.
/*!
@param bands - the bands.
*/
void PrintBands(std::vector<std::vector<double> > const & bands);

//! Make a parametric surface as the source.
/*!
@param The polyData source.
*/
void MakeParametricSurface(vtkPolyData *src);

//! Calculate curvatures.
/*!
The source must be triangulated.
@param src - the source.
@param curv - vtkPolyData with normals and scalar data representing curvatures.
*/
void CalculateCurvatures(vtkPolyData *src, vtkPolyData *curv);

/*!
@param lut - An indexed lookup table.
*/
void MakeLUT(vtkLookupTable *lut);

//! Create a lookup table with the colors reversed.
/*!
@param lut - An indexed lookup table.
@param lutr - The reversed indexed lookup table.
*/
void ReverseLUT(vtkLookupTable *lut, vtkLookupTable *lutr);

//! Count the number of scalars in each band.
/*!
@param bands - the bands.
@param src - the vtkPolyData source.
@return The frequencies of the scalars in each band.
*/
std::vector<int> Frequencies(
  std::vector<std::vector<double> > const & bands,
  vtkPolyData * src);

//! Print the frequency table.
/*!
@param freq - the frequencies.
*/
void PrintFrequencies(std::vector<int> & freq);

//!  Glyph the normals on the surface.
/*!
@param src - the vtkPolyData source.
@param reverseNormals - if True the normals on the surface are reversed.
@param glyph - The glyphs.
*/
void MakeGlyphs(vtkPolyData *src, bool const & reverseNormals,
  vtkGlyph3D *glyph);

//! Assemble the surface for display.
/*!
@param st - the surface to display.
@param iren - the interactor.
*/
void Display(vtkRenderWindowInteractor *iren);

//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
std::vector<std::vector<double> > MakeBandsManually()
  {
  std::vector<std::vector<double> > bands;
  std::vector<double> b(3,0);
  b[0] = -1.0; b[2] = -0.8; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 0; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 0.2; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 0.5; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 1.0; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 1.0; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 23; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 28; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 30; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  b[0] = b[2]; b[2] = 37; b[1] = b[0] + (b[2] - b[0]) / 2.0;
  bands.push_back(b);
  return bands;
  }

//-----------------------------------------------------------------------------
std::vector<std::vector<double> > MakeBands(double const dR[2],
  int const & numberOfBands, bool const & nearestInteger)
  {
  std::vector<std::vector<double> > bands;
  if ((dR[1] < dR[0]) || (numberOfBands <= 0))
    {
    return bands;
    }
  double x[2];
  for (int i = 0; i < 2; ++i)
    {
    x[i] = dR[i];
    }
  if (nearestInteger)
    {
    x[0] = std::floor(x[0]);
    x[1] = std::ceil(x[1]);
    }
  double dx = (x[1] - x[0]) / static_cast<double>(numberOfBands);
  std::vector<double> b;
  b.push_back(x[0]);
  b.push_back(x[0] + dx / 2.0);
  b.push_back(x[0] + dx);
  for (int i = 0; i < numberOfBands; ++i)
    {
    bands.push_back(b);
    for (std::vector<double>::iterator p = b.begin(); p != b.end(); ++p)
      {
      *p = *p + dx;
      }
    }
  return bands;
  }

//-----------------------------------------------------------------------------
std::vector<std::vector<double> > MakeIntegralBands(double const dR[2])
{
  std::vector<std::vector<double> > bands;
  if ( dR[1] < dR[0] )
    {
    return bands;
    }
  double x[2];
  for (int i = 0; i < 2; ++i)
    {
    x[i] = dR[i];
    }
  x[0] = std::floor(x[0]);
  x[1] = std::ceil(x[1]);
  int numberOfBands = static_cast<int>(std::abs(x[1]) + std::abs(x[0]));
  return MakeBands(x,numberOfBands, false);
}

//-----------------------------------------------------------------------------
void PrintBands(std::vector<std::vector<double> > const & bands)
{
  STLHelpers stlHelpers = STLHelpers();
  for (std::vector<std::vector<double> >::const_iterator p = bands.begin();
        p != bands.end(); ++p)
    {
    if ( p == bands.begin() )
      {
      std::cout << "[";
      }
    for (std::vector<double>::const_iterator q = p->begin(); q != p->end(); ++q)
      {
        if ( q == p->begin() )
          {
          std::cout << "[";
          }
        if ( !stlHelpers.IsLast(q,*p) )
          {
          std::cout << *q << ", ";
          }
        else
          {
          std::cout << *q << "]";
          }
      }
    if ( !stlHelpers.IsLast(p,bands) )
      {
      std::cout << ", ";
      }
    else
      {
      std::cout << "]";
      }
    }
  std::cout << std::endl;
}

//-----------------------------------------------------------------------------
void MakeParametricSurface(vtkPolyData *src)
{
  VTK_SP(vtkParametricRandomHills, fn);
  fn->AllowRandomGenerationOn();
  fn->SetRandomSeed(1);
  fn->SetNumberOfHills(30);

  VTK_SP(vtkParametricFunctionSource, source);
  source->SetParametricFunction(fn);
  source->SetUResolution(50);
  source->SetVResolution(50);
  source->SetScalarModeToZ();
  source->Update();
  // Name the arrays (not needed in VTK 6.2+ for vtkParametricFunctionSource)
  source->GetOutput()->GetPointData()->GetNormals()->SetName("Normals");
  source->GetOutput()->GetPointData()->GetScalars()->SetName("Scalars");
  src->DeepCopy(source->GetOutput());
  CalculateCurvatures(source->GetOutput(), src);
}

//-----------------------------------------------------------------------------
void CalculateCurvatures(vtkPolyData *src, vtkPolyData *curv)
{
  // Calculate the curvature.
  VTK_SP(vtkCurvatures, curvature);
  curvature->SetCurvatureTypeToGaussian();
  curvature->SetInputData(src);
  curvature->Update();

  // We will clip the random hills because gaussian curvatures on the
  // edges may fluctuate wildly (especially at corners).
  // We clip with an implicit function made up of planes.
  double dx = 0.5;
  double dy = 0.5;
  double bounds[6] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
  curvature->GetOutput()->GetBounds(bounds);

  VTK_SP(vtkPlane, plane1);
  plane1->SetOrigin(bounds[0] + dx, 0, 0);
  plane1->SetNormal(1, 0, 0);

  VTK_SP(vtkPlane, plane2);
  plane2->SetOrigin(bounds[1] - dx, 0, 0);
  plane2->SetNormal(-1, 0, 0);

  VTK_SP(vtkPlane, plane3);
  plane3->SetOrigin(0, bounds[2] + dy, 0);
  plane3->SetNormal(0, 1, 0);

  VTK_SP(vtkPlane, plane4);
  plane4->SetOrigin(0, bounds[3] - dy, 0);
  plane4->SetNormal(0, -1, 0);

  VTK_SP(vtkImplicitBoolean, clipFunction);
  clipFunction->SetOperationTypeToUnion();
  clipFunction->AddFunction(plane1);
  clipFunction->AddFunction(plane2);
  clipFunction->AddFunction(plane3);
  clipFunction->AddFunction(plane4);

  // Create clipper for the random hills
  VTK_SP(vtkClipPolyData, clipper);
  clipper->SetClipFunction(clipFunction);
  clipper->SetInputConnection(curvature->GetOutputPort());
  clipper->GenerateClipScalarsOff();
  clipper->GenerateClippedOutputOff();
  //clipper->GenerateClippedOutputOn();
  clipper->Update();
  curv->DeepCopy(clipper->GetOutput());
}

//-----------------------------------------------------------------------------
void MakeLUT(vtkLookupTable *lut)
{
  // Make the lookup table.
  VTK_SP(vtkColorSeries, colorSeries);
  // Select a color scheme.
  int colorSeriesEnum;
  //colorSeriesEnum = colorSeries->BREWER_DIVERGING_BROWN_BLUE_GREEN_9;
  colorSeriesEnum = colorSeries->BREWER_DIVERGING_SPECTRAL_10;
  //colorSeriesEnum = colorSeries->BREWER_DIVERGING_PURPLE_ORANGE_9;
  //colorSeriesEnum = colorSeries->BREWER_SEQUENTIAL_BLUE_PURPLE_9;
  //colorSeriesEnum = colorSeries->BREWER_SEQUENTIAL_BLUE_GREEN_9;
  //colorSeriesEnum = colorSeries->BREWER_QUALITATIVE_SET3;
  //colorSeriesEnum = colorSeries->CITRUS;
  colorSeries->SetColorScheme(colorSeriesEnum);

  colorSeries->BuildLookupTable(lut);
  lut->SetNanColor(1,0,0,1);
}

//! Create a lookup table with the colors reversed.
/*!
@param lut - An indexed lookup table.
@param lutr - The reversed indexed lookup table.
*/
//-----------------------------------------------------------------------------
void ReverseLUT(vtkLookupTable *lut, vtkLookupTable *lutr)
{
  // First do a deep copy just to get the whole structure
  // and then reverse the colors and annotations.
  lutr->DeepCopy(lut);
  vtkIdType t = lut->GetNumberOfTableValues() - 1;
  for (vtkIdType i = t; i >= 0; --i)
    {
    double rgba[4] = { 0.0, 0.0, 0.0, 0.0 };
    lut->GetColor(i, rgba);
    rgba[3] = lut->GetOpacity(i);
    lutr->SetTableValue(t - i, rgba);
    }
  t = lut->GetNumberOfAnnotatedValues() - 1;
  for (vtkIdType i = t; i >= 0; --i)
    {
    lutr->SetAnnotation(t - i, lut->GetAnnotation(i));
    }
  }

//-----------------------------------------------------------------------------
std::vector<int> Frequencies(
      std::vector<std::vector<double> > const & bands,
      vtkPolyData * src)
{
  std::vector<int> freq(bands.size(),0);
  vtkIdType tuples =
    src->GetPointData()->GetScalars()->GetNumberOfTuples();
  for (int i = 0; i < tuples; ++i)
    {
    double *x = src->GetPointData()->GetScalars()->GetTuple(i);
    for ( size_t j = 0; j < bands.size(); ++j )
      {
        if (*x <= bands[j][2])
          {
          freq[j] = freq[j] + 1;
          break;
          }
      }
    }
  return freq;
}

//-----------------------------------------------------------------------------
void PrintFrequencies(std::vector<int> & freq)
{
  STLHelpers stlHelpers = STLHelpers();
  int i = 0;
  for (std::vector<int>::const_iterator p = freq.begin(); p != freq.end(); ++p)
    {
    if (p == freq.begin())
      {
      std::cout << "[";
      }
    if (stlHelpers.IsLast(p, freq))
      {
      std::cout << i << ": " << *p << "]";
      }
    else
      {
      std::cout << i << ": " << *p << ", ";
      }
    ++i;
    }
  std::cout << endl;
}

//-----------------------------------------------------------------------------
void MakeGlyphs(vtkPolyData *src, bool const & reverseNormals, vtkGlyph3D *glyph)
{
  // Sometimes the contouring algorithm can create a volume whose gradient
  // vector and ordering of polygon(using the right hand rule) are
  // inconsistent. vtkReverseSense cures this problem.
  VTK_SP(vtkReverseSense, reverse);
  VTK_SP(vtkMaskPoints, maskPts);
  maskPts->SetOnRatio(5);
  maskPts->RandomModeOn();
  if (reverseNormals)
    {
    reverse->SetInputData(src);
    reverse->ReverseCellsOn();
    reverse->ReverseNormalsOn();
    maskPts->SetInputConnection(reverse->GetOutputPort());
    }
  else
    {
    maskPts->SetInputData(src);
    }

  // Source for the glyph filter
  VTK_SP(vtkArrowSource, arrow);
  arrow->SetTipResolution(16);
  arrow->SetTipLength(0.3);
  arrow->SetTipRadius(0.1);

  glyph->SetSourceConnection(arrow->GetOutputPort());
  glyph->SetInputConnection(maskPts->GetOutputPort());
  glyph->SetVectorModeToUseNormal();
  glyph->SetScaleFactor(1);
  glyph->SetColorModeToColorByVector();
  glyph->SetScaleModeToScaleByVector();
  glyph->OrientOn();
  glyph->Update();
}

//-----------------------------------------------------------------------------
//! Make and display the surface.
void Display(vtkRenderWindowInteractor *iren)
{
  // ------------------------------------------------------------
  // Create the surface, lookup tables, contour filter etc.
  // ------------------------------------------------------------
  VTK_SP(vtkPolyData, src);
  MakeParametricSurface(src);
  // The sacalars are named "Scalars"by default
  // in the parametric surfaces, so change the name.
  src->GetPointData()->GetScalars()->SetName("Elevation");
  double scalarRange[2];
  src->GetScalarRange(scalarRange);
  std::cout << "Scalar range: (" << scalarRange[0] << ", " << scalarRange[1] << ")" << std::endl;

  // Let's manually look at the scalars and create a frequency table.
  // We will use this to decide how to manually definr the lookup table.
  std::vector<std::vector<double> > bands1 =
    MakeBands(scalarRange, 20, false);
  PrintBands(bands1);
  // Let's do a frequency table.
  PrintFrequencies(Frequencies(bands1, src));
  std::vector<std::vector<double> > bands =
    MakeBandsManually();
  PrintBands(bands);
  // Let's do a frequency table.
  PrintFrequencies(Frequencies(bands, src));

  VTK_SP(vtkLookupTable, lut);
  MakeLUT(lut);
  lut->SetTableRange(scalarRange);
  vtkIdType numberOfBands = lut->GetNumberOfTableValues();

  //std::vector<std::vector<double> > bands =
  //  MakeBands(scalarRange, numberOfBands, false);
  //PrintBands(bands);

  //// Let's do a frequency table.
  //// The number of scalars in each band.
  //std::vector<int> freq = Frequencies(bands, src);
  //PrintFrequencies(freq);

  // We will use the midpoint of the band as the label.
  std::vector<std::string> labels;
  for (std::vector<std::vector<double> >::const_iterator p = bands.begin();
    p != bands.end(); ++p)
    {
    std::ostringstream os;
    os << std::fixed << std::setw(6) << std::setprecision(2) << (*p)[1];
    labels.push_back(os.str());
    }

  // Annotate
  VTK_SP(vtkVariantArray, values);
  for (size_t i = 0; i < labels.size(); ++i)
    {
    values->InsertNextValue(vtkVariant(labels[i]));
    }
  for (vtkIdType i = 0; i < values->GetNumberOfTuples(); ++i)
    {
    lut->SetAnnotation(i, values->GetValue(i).ToString());
    }

  // Create a lookup table with the colors reversed.
  VTK_SP(vtkLookupTable, lutr);
  ReverseLUT(lut, lutr);

  // Create the contour bands.
  VTK_SP(vtkBandedPolyDataContourFilter, bcf);
  bcf->SetInputData(src);
  // Use either the minimum or maximum value for each band.
  int i = 0;
  for (std::vector<std::vector<double> >::const_iterator p = bands.begin();
    p != bands.end(); ++p)
    {
    bcf->SetValue(i, (*p)[2]);
    ++i;
    }
  // We will use an indexed lookup table.
  bcf->SetScalarModeToIndex();
  bcf->GenerateContourEdgesOn();

  // Generate the glyphs on the original surface.
  VTK_SP(vtkGlyph3D, glyph);
  MakeGlyphs(src, true, glyph);

  // ------------------------------------------------------------
  // Create the mappers and actors
  // ------------------------------------------------------------

  VTK_SP(vtkPolyDataMapper, srcMapper);
  srcMapper->SetInputConnection(bcf->GetOutputPort());
  srcMapper->SetScalarRange(scalarRange);
  srcMapper->SetLookupTable(lut);
  srcMapper->SetScalarModeToUseCellData();

  VTK_SP(vtkActor, srcActor);
  srcActor->SetMapper(srcMapper);
  srcActor->RotateX(-45);
  srcActor->RotateZ(45);

  // Create contour edges
  VTK_SP(vtkPolyDataMapper, edgeMapper);
  edgeMapper->SetInputData(bcf->GetContourEdgesOutput());
  edgeMapper->SetResolveCoincidentTopologyToPolygonOffset();

  VTK_SP(vtkActor, edgeActor);
  edgeActor->SetMapper(edgeMapper);
  edgeActor->GetProperty()->SetColor(0, 0, 0);
  edgeActor->RotateX(-45);
  edgeActor->RotateZ(45);

  VTK_SP(vtkPolyDataMapper, glyphMapper);
  glyphMapper->SetInputConnection(glyph->GetOutputPort());
  glyphMapper->SetScalarModeToUsePointFieldData();
  glyphMapper->SetColorModeToMapScalars();
  glyphMapper->ScalarVisibilityOn();
  glyphMapper->SelectColorArray("Elevation");
  // Colour by scalars.
  // The default lookup table is used but you can
  // use whatever lookup table you like.
  glyphMapper->SetScalarRange(scalarRange);

  VTK_SP(vtkActor, glyphActor);
  glyphActor->SetMapper(glyphMapper);
  glyphActor->RotateX(-45);
  glyphActor->RotateZ(45);

  // Add a scalar bar->
  VTK_SP(vtkScalarBarActor, scalarBar);
  // This LUT puts the lowest value at the top of the scalar bar.
  //scalarBar->SetLookupTable(lut);
  // Use this LUT if you want the highest value at the top.
  scalarBar->SetLookupTable(lutr);
  scalarBar->SetTitle("Elevation (m)");

  // ------------------------------------------------------------
  // Create the RenderWindow, Renderer and Interactor
  // ------------------------------------------------------------
  VTK_SP(vtkRenderer, ren);
  VTK_SP(vtkRenderWindow, renWin);

  renWin->AddRenderer(ren);
  iren->SetRenderWindow(renWin);

  // add actors
  ren->AddViewProp(srcActor);
  ren->AddViewProp(edgeActor);
  ren->AddViewProp(glyphActor);
  ren->AddActor2D(scalarBar);

  ren->SetBackground(0.7, 0.8, 1.0);
  renWin->SetSize(800, 800);
  renWin->Render();

  ren->GetActiveCamera()->Zoom(1.5);
}

//-----------------------------------------------------------------------------
//! Make and display the surface.
int main(int argc, char *argv[])
{
  VTK_SP(vtkRenderWindowInteractor, iren);
  Display(iren);
  iren->Render();
  iren->Start();

  return EXIT_SUCCESS;
}
